""" .. A script containing all relevant functions for the crossover process. """

import numpy as np
import pandas as pd

from ..individual import Individual
from .util import _add_line, _fillna, _get_pdf_counts


def _collate_parents(parent1, parent2):
    """ Concatenate the columns and metadata from each parent together. These
    lists form a pool from which information is inherited during the crossover
    process. """

    parent_columns, parent_metadata = [], []
    for dataframe, metadata in [parent1, parent2]:
        parent_columns += [dataframe[col] for col in dataframe.columns]
        parent_metadata += metadata

    return parent_columns, parent_metadata


def _cross_minimum_cols(parent_columns, parent_metadata, col_limits, pdfs):
    """ If :code:`col_limits` has a tuple lower limit, inherit the minimum
    number of columns from two parents to satisfy this limit. Return part of a
    whole individual and the adjusted parent information. """

    all_idxs, cols, metadata = [], [], []
    for limit, pdf_class in zip(col_limits[0], pdfs):
        if limit:
            pdf_class_idxs = np.where(
                [isinstance(pdf, pdf_class) for pdf in parent_metadata]
            )

            idxs = np.random.choice(*pdf_class_idxs, size=limit, replace=False)
            for idx in idxs:
                metadata.append(parent_metadata[idx])
                cols.append(parent_columns[idx])
                all_idxs.append(idx)

    for idx in sorted(all_idxs, reverse=True):
        parent_columns.pop(idx)
        parent_metadata.pop(idx)

    return cols, metadata, parent_columns, parent_metadata


def _cross_remaining_cols(
    cols, metadata, ncols, parent_columns, parent_metadata, col_limits, pdfs
):
    """ Regardless of whether :code:`col_limits` has a tuple upper limit or not,
    inherit all remaining columns from the two parents so as not to exceed these
    bounds. Return the components of a full inidividual. """

    pdf_counts = _get_pdf_counts(metadata, pdfs)
    while len(cols) < ncols:
        idx = np.random.randint(len(parent_columns))
        pdf = parent_metadata[idx]
        pdf_idx = pdfs.index(pdf.__class__)

        try:
            if pdf_counts[pdf.__class__] < col_limits[1][pdf_idx]:
                cols.append(parent_columns.pop(idx))
                metadata.append(parent_metadata.pop(idx))
                pdf_counts[pdf.__class__] += 1

        except TypeError:
            cols.append(parent_columns.pop(idx))
            metadata.append(parent_metadata.pop(idx))

    return cols, metadata


def crossover(parent1, parent2, col_limits, pdfs, prob=0.5):
    """ Blend the information from two parents to create a new
    :code:`Individual`. Dimensions are inherited first, and then column-metadata
    pairs are inherited from either parent uniformly. Missing values are filled
    in as necessary.

    Parameters
    ----------
    parent1 : Individual
        The first individual to be blended.
    parent2 : Individual
        The second individual to be blended.
    col_limits : list
        Lower and upper bounds on the number of columns :code:`offspring` can
        have. Used in case of tuple limits.
    pdfs : list
        Families of distributions with which to create new columns. Used in case
        of tuple column limits.
    prob : float, optional
        The cut-off probability with which to inherit dimensions from
        :code:`parent1` over :code:`parent2`.

    Returns
    -------
    offspring : Individual
        A new individual formed from the dimensions and columns of its parents.
    """

    parent_columns, parent_metadata = _collate_parents(parent1, parent2)
    cols, metadata = [], []

    if np.random.random() < prob:
        nrows = len(parent1.dataframe)
    else:
        nrows = len(parent2.dataframe)

    if np.random.random() < prob:
        ncols = len(parent1.metadata)
    else:
        ncols = len(parent2.metadata)

    if isinstance(col_limits[0], tuple):
        cols, metadata, parent_columns, parent_metadata = _cross_minimum_cols(
            parent_columns, parent_metadata, col_limits, pdfs
        )

    cols, metadata = _cross_remaining_cols(
        cols, metadata, ncols, parent_columns, parent_metadata, col_limits, pdfs
    )

    dataframe = pd.DataFrame({i: col for i, col in enumerate(cols)})

    while len(dataframe) != nrows:
        if len(dataframe) > nrows:
            dataframe = dataframe.iloc[:nrows, :]
        elif len(dataframe) < nrows:
            dataframe, metadata = _add_line(dataframe, metadata, axis=0)

    dataframe = _fillna(dataframe, metadata)
    return Individual(dataframe, metadata)



""" A script containing functions for each of the components of the genetic
algorithm. """

import numpy as np

from .individual import create_individual
from .operators import crossover, mutation


def create_initial_population(size, row_limits, col_limits, pdfs, weights=None):
    """ Create an initial population for the genetic algorithm based on the
    given parameters.

    Parameters
    ----------
    size : int
        The number of individuals in the population.
    row_limits : list
        Limits on the number of rows a dataset can have.
    col_limits : list
        Limits on the number of columns a dataset can have.
    pdfs : list
        A list of potential column pdf classes such as those found in
        `pdfs.py`. Must have a `.sample()` and `.mutate()` method.
    weights : list
        A sequence of relative weights the same length as `column_classes`. This
        acts as a loose probability distribution from which to sample column
        classes. If `None`, column classes are sampled equally.

    Returns
    -------
    population : list
        A collection of individuals.
    """

    if size <= 1:
        raise ValueError(
            "There must be more than one individual in a \
                          population"
        )

    population = [
        create_individual(row_limits, col_limits, pdfs, weights)
        for _ in range(size)
    ]

    return population


def create_new_population(
    parents,
    size,
    crossover_prob,
    mutation_prob,
    row_limits,
    col_limits,
    pdfs,
    weights,
):
    """ Given a set of potential parents to be carried into the next generation,
    create offspring from pairs within that set until there are enough
    individuals. Each individual offspring is formed using a crossover operator
    on the two parent individuals and then mutating them according to the
    probability `mutation_prob`. """

    population = parents
    while len(population) < size:
        parent1_idx, parent2_idx = np.random.choice(len(parents), size=2)
        parent1, parent2 = parents[parent1_idx], parents[parent2_idx]
        offspring = crossover(
            parent1, parent2, col_limits, pdfs, crossover_prob
        )
        mutant = mutation(
            offspring, mutation_prob, row_limits, col_limits, pdfs, weights
        )
        population.append(mutant)

    return population



""" All currently implemented, continuous distributions. """

import numpy as np

from .base import Distribution


class Gamma(Distribution):
    """ Continuous column class given by the gamma distribution.

    Attributes
    ----------
    name : str
        Name of the distribution, :code:`"Gamma"`.
    dtype : str
        Preferred datatype of distribution, :code:`"float"`.
    param_limits : dict
        A dictionary of limits on the distribution parameters. Defaults to
        :code:`[0, 10]` for both :code:`alpha` and :code:`theta`.
    alpha : float
        The shape parameter sampled from :code:`param_limits["alpha"]`.
        `Instance attribute.`
    theta : float
        The scale parameter sampled from :code:`param_limits["theta"]`.
        `Instance attribute.`
    """

    name = "Gamma"
    dtype = "float"
    hard_limits = {"alpha": [0, 10], "theta": [0, 10]}
    param_limits = {"alpha": [0, 10], "theta": [0, 10]}

    def __init__(self):

        self.alpha = np.random.uniform(*self.param_limits["alpha"])
        self.theta = np.random.uniform(*self.param_limits["theta"])

    def sample(self, nrows):
        """ Take a sample of size :code:`nrows` from the gamma distribution with
        shape and scale parameters given by :code:`alpha` and :code:`theta`
        respectively. """

        return np.random.gamma(shape=self.alpha, scale=self.theta, size=nrows)


class Normal(Distribution):
    """ Continuous column class given by the normal distribution.

    Attributes
    ----------
    name : str
        Name of the distribution, :code:`"Normal"`.
    dtype : str
        Preferred datatype of distribution, :code:`"float"`.
    param_limits : dict
        A dictionary of limits on the distribution parameters. Defaults to
        :code:`[-10, 10]` for :code:`mean` and :code:`[0, 10]` for :code:`std`.
    mean : float
        The mean, sampled from :code:`param_limits["mean"]`. `Instance
        attribute.`
    std : float
        The standard deviation, sampled from :code:`param_limits["std"]`.
        `Instance attribute.`
    """

    name = "Normal"
    dtype = "float"
    hard_limits = {"mean": [-10, 10], "std": [0, 10]}
    param_limits = {"mean": [-10, 10], "std": [0, 10]}

    def __init__(self):

        self.mean = np.random.uniform(*self.param_limits["mean"])
        self.std = np.random.uniform(*self.param_limits["std"])

    def sample(self, nrows):
        """ Take a sample of size :code:`nrows` from the normal distribution
        with mean and standard deviation given by :code:`mean` and :code:`std`
        respectively. """

        return np.random.normal(loc=self.mean, scale=self.std, size=nrows)



""" .. Base inheritance class for all distributions. """

from copy import deepcopy


class Distribution:
    """ A base class for all currently implemented distributions and those
    defined by users.

    Attributes
    ----------
    name : str
        The name of the distribution, "Distribution".
    hard_limits : None
        A placeholder for a dictionary with hard bounds the parameters. Also
        acts as the default limits when resetting an individual.
    param_limits : None
        A placeholder for a distribution parameter limit dictionary.
    """

    name = "Distribution"
    hard_limits = None
    param_limits = None

    def __repr__(self):

        params = ""
        for key, val in self.__dict__.items():
            params += f"{key}={val:.2f}, "
        params = params[:-2]

        return f"{self.name}({params})"

    @classmethod
    def reset(cls):
        """ Reset the class to have its widest parameter limits, given in the
        class attribute :code:`hard_limits`. """

        cls.param_limits = deepcopy(cls.hard_limits)

    def sample(self):
        """ Raise a :code:`NotImplementedError` by default. """

        raise NotImplementedError("You must define a sample method.")

    def to_tuple(self):
        """ Returns the name of distribution, and the names and values of all
        parameters as a tuple. This is used for the saving of data and little
        else. """

        out = [self.name]
        for key, val in self.__dict__.items():
            out.append(key)
            out.append(val)

        return tuple(out)



""" .. Functions related to the mutation operator. """

from copy import deepcopy

import numpy as np

from ..individual import Individual
from .util import _add_line, _remove_line


def _mutate_nrows(dataframe, metadata, row_limits, prob):
    """ Mutate the number of rows an individual has by adding a new row and/or
    dropping a row at random so as not to exceed the bounds of
    :code:`row_limits`. """

    if np.random.random() < prob and dataframe.shape[0] < row_limits[1]:
        dataframe, metadata = _add_line(dataframe, metadata, axis=0)

    if np.random.random() < prob and dataframe.shape[0] > row_limits[0]:
        dataframe, metadata = _remove_line(dataframe, metadata, axis=0)

    return dataframe, metadata


def _mutate_ncols(dataframe, metadata, col_limits, pdfs, weights, prob):
    """ Mutate the number of columns an individual has by adding a new column
    and/or dropping a column at random. In either case, the bounds defined in
    :code:`col_limits` cannot be exceeded. """

    if isinstance(col_limits[1], tuple):
        condition = dataframe.shape[1] < sum(col_limits[1])
    else:
        condition = dataframe.shape[1] < col_limits[1]

    if np.random.random() < prob and condition:
        dataframe, metadata = _add_line(
            dataframe, metadata, 1, col_limits, pdfs, weights
        )

    if isinstance(col_limits[0], tuple):
        condition = dataframe.shape[1] > sum(col_limits[0])
    else:
        condition = dataframe.shape[1] > col_limits[0]

    if np.random.random() < prob and condition:
        dataframe, metadata = _remove_line(
            dataframe, metadata, 1, col_limits, pdfs
        )

    return dataframe, metadata


def _mutate_values(dataframe, metadata, prob):
    """ Iterate over the values of :code:`dataframe`, mutating them with
    probability :code:`prob`. Mutation is done by resampling from each column's
    associated distribution in :code:`metadata`. """

    for j, col in enumerate(dataframe.columns):
        pdf = metadata[j]
        for i, value in enumerate(dataframe[col]):
            if np.random.random() < prob:
                value = pdf.sample(1)[0]
                dataframe.iloc[i, j] = value

    return dataframe


def mutation(individual, prob, row_limits, col_limits, pdfs, weights=None):
    """ Mutate an individual. Here, the characteristics of an individual can be
    split into two parts: their dimensions, and their values. Each of these
    parts is mutated in a different way using the same probability,
    :code:`prob`.

    Parameters
    ----------
    individual : Individual
        The individual to be mutated.
    prob : float
        The probability with which any characteristic of :code:`individual`
        should be mutated.
    row_limits : list
        Lower and upper limits on the number of rows an individual can have.
    col_limits : list
        Lower and upper limits on the number of columns an individual can have.
    pdfs : list
        Families of distributions with which to create new columns.
    weights : list, optional
        Probability with which to sample a distribution from :code:`pdfs`. If
        :code:`None`, sample uniformly.

    Returns
    -------
    mutant : Individual
        A (potentially) mutated individual.
    """

    dataframe, metadata = deepcopy(individual)
    dataframe, metadata = _mutate_nrows(dataframe, metadata, row_limits, prob)
    dataframe, metadata = _mutate_ncols(
        dataframe, metadata, col_limits, pdfs, weights, prob
    )

    dataframe = _mutate_values(dataframe, metadata, prob)
    return Individual(dataframe, metadata)



""" A collection of objects related to the definition and creation of an
individual in this GA. An individual is defined by a dataframe and its
associated metadata. This metadata is simply a list of the distributions from
which each column of the dataframe was generated. These are reused during
mutation and for filling in missing values during crossover. """

from collections import namedtuple

import numpy as np
import pandas as pd

Individual = namedtuple("Individual", ["dataframe", "metadata"])


def _sample_ncols(col_limits):
    """ Sample a valid number of columns from the column limits, even if those
    limits contain tuples. """

    integer_limits = []
    for lim in col_limits:
        try:
            integer_lim = sum(lim)
        except TypeError:
            integer_lim = lim
        integer_limits.append(integer_lim)

    return np.random.randint(integer_limits[0], integer_limits[1] + 1)


def _get_minimum_cols(cols, metadata, nrows, col_limits, pdfs, pdf_counts):
    """ If :code:`col_limits` has a tuple lower limit then sample columns of the
    correct class from :code:`pdfs` as needed to satisfy this bound. """

    for i, min_limit in enumerate(col_limits[0]):
        pdf_class = pdfs[i]
        for _ in range(min_limit):
            pdf = pdf_class()
            cols.append(pdf.sample(nrows))
            metadata.append(pdf)
            pdf_counts[pdf_class] += 1

    return cols, metadata, pdf_counts


def _get_remaining_cols(
    cols, metadata, nrows, ncols, col_limits, pdfs, weights, pdf_counts
):
    """ Sample all remaining columns for the current individual. If
    :code:`col_limits` has a tuple upper limit then sample all remaining
    columns for the individual without exceeding the bounds. """

    while len(cols) < ncols:
        pdf_class = np.random.choice(pdfs, p=weights)
        pdf = pdf_class()
        try:
            idx = pdfs.index(pdf_class)
            if pdf_counts[pdf_class] < col_limits[1][idx]:
                cols.append(pdf.sample(nrows))
                metadata.append(pdf)
                pdf_counts[pdf_class] += 1
        except TypeError:
            cols.append(pdf.sample(nrows))
            metadata.append(pdf)

    return cols, metadata


def create_individual(row_limits, col_limits, pdfs, weights=None):
    """ Create an individual dataset-metadata representation within the limits
    provided. An individual is contained within a :code:`namedtuple` object.

    Parameters
    ----------
    row_limits : list
        Lower and upper bounds on the number of rows a dataset can have.
    col_limits : list
        Lower and upper bounds on the number of columns a dataset can have.
        Tuples can be used to indicate limits on the number of columns needed to
    pdfs : list
        A list of potential column pdf classes to select from such as those
        found in :code:`genetic_data.pdfs`.
    weights : list
        A sequence of relative weights the same length as :code:`pdfs`. This
        acts as a probability distribution from which to sample column classes.
        If :code:`None`, column classes are sampled uniformly.
    """

    nrows = np.random.randint(row_limits[0], row_limits[1] + 1)
    ncols = _sample_ncols(col_limits)

    cols, metadata = [], []
    pdf_counts = {pdf_class: 0 for pdf_class in pdfs}

    if isinstance(col_limits[0], tuple):
        cols, metadata, pdf_counts = _get_minimum_cols(
            cols, metadata, nrows, col_limits, pdfs, pdf_counts
        )

    cols, metadata = _get_remaining_cols(
        cols, metadata, nrows, ncols, col_limits, pdfs, weights, pdf_counts
    )

    dataframe = pd.DataFrame({i: col for i, col in enumerate(cols)})
    return Individual(dataframe, metadata)



""" Fitness-related functions. """


def get_fitness(fitness, population, fitness_kwargs=None):
    """ Return the fitness score of each individual in a population. """

    if fitness_kwargs:
        return [
            fitness(individual.dataframe, **fitness_kwargs)
            for individual in population
        ]

    return [fitness(individual.dataframe) for individual in population]
